from typing import TextIO
from ruamel import yaml
from .serialisation import TextSerialisation


class YamlSerialisation(TextSerialisation, register="yaml"):
    EXTENSION = "yaml"

    def dump(self, datastructure: dict, textstream: TextIO):
        yaml.dump(datastructure, textstream, yaml.Dumper)

    def load(self, textstream: TextIO) -> dict:
        return yaml.load(textstream, yaml.Loader)



from .stanza_handler import TypedStanzaHandler


class ListStanzaHandler(TypedStanzaHandler, register="list"):
    @property
    def _type(self):
        return list

    def _merge(self, stanza_data: list, new_data: list):
        stanza_data += new_data



from abc import abstractmethod
from typing import BinaryIO
from plugable import Plugable


class Filesystem(Plugable):
    def __init__(self):
        self._history = []

    @abstractmethod
    def get_filename(self, path, base=None) -> str:
        raise NotImplementedError

    @abstractmethod
    def _open(self, filename: str) -> BinaryIO:
        raise NotImplementedError

    @abstractmethod
    def exists(self, filename: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_dir_name(self, filename) -> str:
        raise NotImplementedError

    def canonicalise(self, relative_filename: str):
        absolute_fn = self.get_filename([relative_filename])
        if self.exists(absolute_fn):
            return absolute_fn
        raise FileNotFoundError(absolute_fn)

    def open(self, filename) -> BinaryIO:
        self._history.append(filename)
        return self._open(filename)

    @property
    def history(self):
        return tuple(self._history)



from .directive import ListDirective
from ..loader_context import LoaderContext


class ImportsDirective(ListDirective, register="imports"):
    def handle_item(self, context: LoaderContext, item: str):
        assert(type(item) == str)
        context.imports[context.current_filename].append(item)



from .stanza_handler import TypedStanzaHandler


class DictStanzaHandler(TypedStanzaHandler, register="dict"):
    @property
    def _type(self):
        return dict

    def _merge(self, stanza_data: dict, new_data: dict):
        stanza_data.update(new_data)



import typing
from abc import abstractmethod
from plugable import Plugable

from ..exceptions import ExclusiveStanzaClashError
from ..loader_context import LoaderContext


class StanzaHandler(Plugable):
    def __init__(self, key: str, exclusive=False):
        self.key = key
        self.exclusive = exclusive
        self.filenames = []
        self._includes = 0

    def _can_merge(self):
        if self.exclusive and self._includes > 1:
            return False
        return True

    def include(self, ctx: LoaderContext, new_data):
        self.filenames.append(ctx.current_filename)
        self._includes += 1
        if self._can_merge():
            if self.key not in ctx.data:
                ctx.data[self.key] = self._initialise()
            self._merge(ctx.data[self.key], new_data)
        else:
            clashing_files = ",".join(self.filenames)
            raise ExclusiveStanzaClashError(self.key, clashing_files)

    @abstractmethod
    def is_valid(self, new_data) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _initialise(self):
        raise NotImplementedError

    @abstractmethod
    def _merge(self, stanza_data: dict, new_data):
        raise NotImplementedError

    def cleanup(self):
        self.filenames = []
        self._includes = 0


class TypedStanzaHandler(StanzaHandler):
    @property
    @abstractmethod
    def _type(self) -> typing.Type:
        raise NotImplementedError

    def _initialise(self):
        return self._type()

    def is_valid(self, new_data):
        if isinstance(new_data, self._type):
            return True
        return False



import re
from copy import copy
from collections import deque

from .exceptions import StanzaValidationError
from .stanzahandlers import StanzaHandler
from .loader_context import LoaderContext
from .serialisations import Serialisation
from .filesystems import Filesystem
from .directives import Directive

DEFAULT_LOADER_WHITELIST = (
    "imports",
)


class FlatLoader:
    _DIRECTIVE_RGX = re.compile(r"\$.+")

    def __init__(
            self,
            serialisation: Serialisation = None,
            directives: tuple = DEFAULT_LOADER_WHITELIST
    ):
        if not serialisation:
            serialisation = Serialisation.get("json")

        self._serialisation = serialisation
        self._stanza_handlers = {}
        self._directives = {name: Directive.get(name) for name in directives}

    def register_handler(self, stanza_handler: StanzaHandler):
        self._stanza_handlers[stanza_handler.key] = stanza_handler

    def set_schema(self, schema: dict):
        self._stanza_handlers = {}
        for key, value in schema.items():
            if isinstance(value, dict):
                value["key"] = key
                handler = StanzaHandler.get(**value)
            else:
                handler = StanzaHandler.get(value, key)

            self.register_handler(handler)

    def _load_file(self, fs: Filesystem, filename: str) -> dict:
        with fs.open(filename) as filestream:
            return self._serialisation.deserialise(filestream)

    def _process_directives(self, ctx: LoaderContext, data: dict) -> dict:
        doc_directives = {
            key: value for key, value in data.items()
            if self._DIRECTIVE_RGX.match(key)
        }

        for key, value in doc_directives.items():
            del data[key]
            key = key[1:]
            if key in self._directives:
                self._directives[key].handle_directive(ctx, value)

        return data

    def _cleanup_handlers(self):
        for handler in self._stanza_handlers.values():
            handler.cleanup()

    def load(self, root_filename: str, fs: Filesystem = None) -> LoaderContext:
        if not fs:
            fs = Filesystem.get("local")

        root_filename = fs.canonicalise(root_filename)
        ctx = LoaderContext()

        stack = deque()
        stack.append(root_filename)

        while stack:
            filename = stack.pop()
            directory = fs.get_dir_name(filename)

            ctx.filenames.append(filename)

            data = self._process_directives(ctx, self._load_file(fs, filename))

            new_imports = copy(ctx.imports[filename])
            new_imports.reverse()

            for new_import in new_imports:
                decoded_import = self._serialisation.decode_import(new_import)
                stack.append(fs.get_filename(decoded_import, directory))

            for key, stanzadata in data.items():
                if key in self._stanza_handlers:
                    handler = self._stanza_handlers[key]

                    if handler.is_valid(stanzadata):
                        handler.include(ctx, stanzadata)
                    else:
                        raise StanzaValidationError(key, filename)

            ctx.import_count += 1
        ctx.imports = dict(ctx.imports)
        self._cleanup_handlers()
        return ctx



import dataclasses

from typing import Dict, List
from collections import defaultdict


def _field(factory):
    return dataclasses.field(default_factory=factory)


@dataclasses.dataclass
class LoaderContext:
    imports: Dict[str, List[str]] = _field(lambda: defaultdict(list))
    filenames: List[str] = _field(list)
    data: dict = _field(dict)
    import_count: int = 0

    @property
    def current_filename(self):
        return self.filenames[self.import_count]



_CLASH_ERR = "Cannot have more than one '{}' stanza (Clashing files: {})"
_VALID_ERR = "The data in the '{}' stanza from '{}' failed validation rules."


class MetaloaderError(RuntimeError):
    pass


class ExclusiveStanzaClashError(MetaloaderError):
    def __init__(self, key, clashing_files):
        error_message = _CLASH_ERR.format(key, clashing_files)
        super().__init__(error_message)


class StanzaValidationError(MetaloaderError):
    def __init__(self, key, filename):
        error_message = _VALID_ERR.format(key, filename)
        super().__init__(error_message)



import metaloader
from integration_test import IntegrationTest

VFS_DATA = {
    "root": {
        "$imports": ["import_1", "import_3"],
        "test_list": [1, 2, 3],
        "test_dict": {
            "4": "4",
            "5": "5",
            "6": "6"
        },
        "test_exclusive": {
            "hello": "world"
        }
    },

    "import_1": {
        "$imports": ["folder.import_2"],
        "test_list": [4, 5, 6],
        "test_dict": {
            "1": "1",
            "2": "2",
            "3": "3"
        }
    },

    "folder.import_2": {
        "test_list": [7, 8, 9],
        "test_dict": {
            "10": "10",
            "11": "11",
            "12": "12"
        }
    },

    "import_3": {
        "test_list": [10, 11, 12],
        "test_dict": {
            "7": "7",
            "8": "8",
            "9": "9"
        }
    },

    "validation_fail": {
        "test_list": {}
    }
}


def test_all_serialisations():
    for serialisation_str in metaloader.Serialisation.registry.enum():
        serialisation = metaloader.Serialisation.get(serialisation_str)
        fs = metaloader.Filesystem.get("virtual", VFS_DATA, serialisation_str)
        test = IntegrationTest(fs, serialisation, serialisation.EXTENSION)
        test.run()



import os
from .filesystem import Filesystem


class LocalFilesystem(Filesystem, register="local"):
    def __init__(self, root: str = None):
        super().__init__()
        if not root:
            root = os.getcwd()
        self._root = root

    def get_dir_name(self, filename):
        return os.path.dirname(filename)

    def get_filename(self, path, base=None):
        if not base:
            base = self._root
        return os.path.join(base, *path)

    def _open(self, filename: str):
        return open(filename, "rb")

    def exists(self, filename: str):
        return os.path.isfile(filename)



import mock
from os.path import join, dirname, abspath
from metaloader import Filesystem, Serialisation
from integration_test import IntegrationTest


def get_absolute_root(root):
    return join(dirname(abspath(__file__)), "data", root)


def mock_cwd(root):
    retval = get_absolute_root(root)
    return mock.patch("os.getcwd", mock.MagicMock(return_value=retval))


def run_local_fs_test(serialisation_str):
    serialisation = None
    fs = None

    if serialisation_str != "json":
        serialisation = Serialisation.get(serialisation_str)
    else:
        fs = Filesystem.get("local", get_absolute_root("json"))

    test = IntegrationTest(fs, serialisation, serialisation_str)
    test.run()

    if fs:
        assert "root.json" in fs.history[0]


def test_json_local_fs():
    run_local_fs_test("json")


@mock_cwd("yaml")
def test_yaml_local_fs():
    run_local_fs_test("yaml")



from io import BufferedReader, BufferedWriter, TextIOWrapper, BytesIO
from abc import abstractmethod
from typing import BinaryIO, TextIO
from plugable import Plugable


class Serialisation(Plugable):
    """
        Represents a method of serialising a data structure
        within a file
    """

    EXTENSION = "dat"

    @classmethod
    def decode_import(cls, import_stmt: str):
        path = import_stmt.split(".")

        path[-1] = ".".join([path[-1], cls.EXTENSION])

        return path

    @abstractmethod
    def serialise(self, datastructure: dict, filestream: BinaryIO):
        raise NotImplementedError

    @abstractmethod
    def deserialise(self, filestream: BinaryIO) -> dict:
        raise NotImplementedError


class TextSerialisation(Serialisation):
    """
        Represents a method of serialising a data structure
        within a text-based file
    """

    def __init__(self, encoding: str = "utf-8"):
        self._encoding = encoding

    @abstractmethod
    def load(self, textstream: TextIO) -> dict:
        raise NotImplementedError

    @abstractmethod
    def dump(self, datastructure: dict, textstream: TextIO):
        raise NotImplementedError

    def serialise(self, datastructure: dict, filestream: BinaryIO) -> BinaryIO:
        buffer = BufferedWriter(filestream)

        kwargs = {
            "buffer": buffer,
            "encoding": self._encoding,
            "write_through": True
        }

        with TextIOWrapper(**kwargs) as textstream:
            self.dump(datastructure, textstream)
            buffer.flush()
            filestream.seek(0)
            return BytesIO(filestream.getvalue())

    def deserialise(self, filestream: BinaryIO):
        buffer = BufferedReader(filestream)
        with TextIOWrapper(buffer, self._encoding) as textstream:
            return self.load(textstream)



from .exceptions import (
    MetaloaderError,
    ExclusiveStanzaClashError,
    StanzaValidationError
)
from .filesystems import Filesystem, LocalFilesystem, VirtualFilesystem
from .serialisations import Serialisation, JsonSerialisation, YamlSerialisation
from .stanzahandlers import StanzaHandler, ListStanzaHandler, DictStanzaHandler
from .loader_context import LoaderContext
from .flat_loader import FlatLoader

__all__ = [
    "MetaloaderError",
    "ExclusiveStanzaClashError",
    "StanzaValidationError",
    "Filesystem",
    "LocalFilesystem",
    "VirtualFilesystem",
    "Serialisation",
    "JsonSerialisation",
    "YamlSerialisation",
    "LoaderContext",
    "StanzaHandler",
    "ListStanzaHandler",
    "DictStanzaHandler",
    "FlatLoader"
]



from abc import abstractmethod
from plugable import Plugable
from ..loader_context import LoaderContext


class Directive(Plugable):
    @abstractmethod
    def handle_directive(self, context: LoaderContext, args):
        raise NotImplementedError


class ListDirective(Directive):
    @abstractmethod
    def handle_item(self, context: LoaderContext, item):
        raise NotImplementedError

    def handle_directive(self, context: LoaderContext, args: list):
        for item in args:
            self.handle_item(context, item)



import json
from typing import TextIO
from .serialisation import TextSerialisation


class JsonSerialisation(TextSerialisation, register="json"):
    EXTENSION = "json"

    def dump(self, datastructure: dict, textstream: TextIO):
        json.dump(datastructure, textstream)

    def load(self, textstream: TextIO) -> dict:
        return json.load(textstream)



import re
from io import BytesIO
from .filesystem import Filesystem
from ..serialisations import Serialisation


PATH_RGX = re.compile(r"(?<=\/).+(?=\..+$)")


class VirtualFilesystem(Filesystem, register="virtual"):
    """
        Represents a virtual filesystem, used for testing
    """

    def __init__(self, data: dict, serialisation: str):
        super().__init__()
        self._data = {}
        self._serialisation = Serialisation.get(serialisation)

        for key, value in data.items():
            path = self._serialisation.decode_import(key)
            filename = self.get_filename(path)
            self._data[filename] = value

    def get_dir_name(self, filename):
        path = "/".join(filename.strip("/").split("/")[:-1])
        return f"/{path}"

    def get_filename(self, path, base=None):
        path = "/".join(path)
        return f"/{path}"

    def _open(self, filename):
        filestream = BytesIO()
        return self._serialisation.serialise(self._data[filename], filestream)

    def exists(self, filename):
        return filename in self._data
